"""
Inventory

This module contains actions for viewing images in the PPA Inventory page.

!!! danger "PPA Version"

    The actions in this module require PPA v2.7.x or newer.
"""
import uuid

from typing import Optional, List, Union

from ppa import ui

from ._api import get_client, create_records, get_cell_value
from .records import Image


TABLE_HEADER = ["Name", "Description", "Author", "Updated At", "Deployed"]
TABLE_FIELDS = ["name", "description", "author", "updated_at", "deployed"]


@create_records(Image)
def get_latest() -> List[Image]:
    """Get the _latest version_ of each inventory image visible to the current user.

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    This action takes no inputs.

??? success "Output"

    A list of [Image](../inputs_outputs/#image) dictionaries.

??? example
    ```YAML linenums="1"
    - ppa_api.inventory.get_latest:
      save: latest_images
    ```
"""
    with get_client() as ppa:
        return ppa.images(latest=True)


@create_records(Image)
def get_all() -> List[Image]:
    """Get _all versions_ of each inventory image visible to the current user.

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    This action takes no inputs.

??? success "Output"

    A list of [Image](../inputs_outputs/#image) dictionaries.

??? example
    ```YAML linenums="1"
    - ppa_api.inventory.get_all:
      save: all_images
    ```
"""
    with get_client() as ppa:
        return ppa.images(latest=False)


@create_records(Image)
def get_deployed() -> List[Image]:
    """Get all _deployed_ inventory images visible to the current user.

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    This action takes no inputs.

??? success "Output"

    A list of [Image](../inputs_outputs/#image) dictionaries.

??? example
    ```YAML linenums="1"
    - ppa_api.inventory.get_deployed:
      save: deployed_images
    ```
"""
    with get_client() as ppa:
        return list(filter(lambda image: image.deployed, ppa.images(latest=True)))


@create_records(Image)
def get_all_undeployed() -> List[Image]:
    """Get all _undeployed_ inventory images visible to the current user.

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    This action takes no inputs.

??? success "Output"

    A list of [Image](../inputs_outputs/#image) dictionaries.

??? example
    ```YAML linenums="1"
    - ppa_api.inventory.get_all_undeployed:
      save: all_undeployed_images
    ```
"""
    with get_client() as ppa:
        return list(filter(lambda image: not image.deployed, ppa.images(latest=False)))


@create_records(Image)
def get_latest_undeployed() -> List[Image]:
    """Get the _latest version_ of each _undeployed_ inventory image visible to the current user.

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    This action takes no inputs.

??? success "Output"

    A list of [Image](../inputs_outputs/#image) dictionaries.

??? example
    ```YAML linenums="1"
    - ppa_api.inventory.get_latest_undeployed:
      save: latest_undeployed_images
    ```
"""
    with get_client() as ppa:
        return list(filter(lambda image: not image.deployed, ppa.images(latest=True)))


@create_records(Image)
def get_latest_by_name(name: str) -> Image:
    """Get the _latest version_ of an image with the supplied name.

The image must be visible to the current user.

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    This action takes no inputs.

??? success "Output"

    Either a single [Image](../inputs_outputs/#image) dictionary or `null`.

??? example
    ```YAML linenums="1"
    - ppa_api.inventory.get_latest_by_name:
        name: Local Group Audit
      save: image
    ```
"""
    with get_client() as ppa:
        return ppa.image_by_name(name)


@create_records(Image)
def get_all_by_name(name: str) -> List[Image]:
    """Get _all versions_ of an image with the supplied name.

The image must be visible to the current user.

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    This action takes no inputs.

??? success "Output"

    Either a list of [Image](../inputs_outputs/#image) dictionaries or `null`.

??? example
    ```YAML linenums="1"
    - ppa_api.inventory.get_all_by_name:
        name: Local Group Audit
      save: images
    ```
"""
    with get_client() as ppa:
        return ppa.images_by_name(name)


def output_table(text: str, images: Union[List[Image], Image]) -> None:
    """Display images in a table.

The table will have the following columns:

- Name
- Description
- Author
- Updated At
- Deployed

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __images:__ a single or list of [Image](../inputs_outputs/#image) dictionaries to display in the table

??? success "Output"

    Nothing is outputted by this action.

??? example
    ```YAML linenums="1"
    - ppa_api.images.get_deployed:
      save: deployed_images

    - ppa_api.images.output_table:
        text: Deployed Images
      load:
        images: deployed_images
    ```
    """
    images = [images] if not isinstance(images, list) else images
    ui.output_table(
        text, header=TABLE_HEADER, data=[[get_cell_value(image, field) for field in TABLE_FIELDS] for image in images]
    )


def input_table(
    text: str, images: Union[List[Image], Image], minimum: Optional[int] = None, maximum: Optional[int] = None
) -> List[Image]:
    """Display images in a table, & allow the task operator to make a selection.

The table will have the following columns:

- Name
- Description
- Author
- Updated At
- Deployed

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __images:__ a single or list of [Image](../inputs_outputs/#image) dictionaries to display in the table

    - __minimum:__ the minimum number of acceptable selections

    - __maximum:__ the maximum number of acceptable selections

??? success "Output"

    A single or list of [Image](../inputs_outputs/#image) dictionaries.

??? tip

    - If neither a _minimum_ or _maximum_ is provided, the task operator will be able to submit 0 selections.
    - If _minimum_ or _maximum_ are provided, the operation will repeat until the task operator makes a valid number
        of selections.

??? example

    ```YAML linenums="1"
    - ppa_api.images.get_deployed:
      save: deployed_images

    - ppa_api.images.input:
        text: Select Image
        minimum: 1
        maximum: 1
      load:
        images: deployed_images
      save: selected_image

    - ppa_api.tasks.start:
      load:
        name: selected_image.name
    ```
    """
    images = [images] if not isinstance(images, list) else images
    table_element_id = str(uuid.uuid4())
    error_element_id = str(uuid.uuid4())
    while True:
        indices = ui.input_table(
            text,
            header=TABLE_HEADER,
            data=[[get_cell_value(image, field) for field in TABLE_FIELDS] for image in images],
            element_id=table_element_id,
        )
        count = len(indices)
        if minimum and count < minimum:
            ui.output_error(f"Please choose at least {minimum} record(s)", element_id=error_element_id)
            continue
        if maximum and count > maximum:
            ui.output_error(f"Please choose no more than {maximum} record(s)", element_id=error_element_id)
            continue
        break

    return [images[i] for i in indices]
