import datetime

from typing import Optional, TypedDict, List, Union, Any, Dict

from datacaster.classes import CastDataClass


DateString = Optional[str]
AnyJson = Union[None, bool, int, float, str, List[Any], Dict[str, Any]]

OUTPUT_DATE_FORMAT = "%H:%M:%S %d/%m/%Y"


def _date_string(date_string: str) -> str:
    # Inbound date strings look like this:
    # 2021-03-23T10:29:05.8413+00:00
    return datetime.datetime.strptime(date_string.split(".")[0], "%Y-%m-%dT%H:%M:%S").strftime(OUTPUT_DATE_FORMAT)


def _optional_date_string(date_string: Optional[str]) -> str:
    return _date_string(date_string) if date_string else None


class Task(TypedDict):
    """
- Contains information about a PPA Task

- Tasks are all based on an Inventory [Image](#image)

!!! example

    ```YAML linenums="1"
    author: John Smith
    cancelled_at:
    cancelled_by:
    duration: '00:00:20.959942'
    exit_code: 0
    exit_message:
    id: 1208
    image: Builtin Administrator Audit
    image_id: 936
    started_at: '08:35:01 24/03/2021'
    state: Success
    stopped_at: '08:35:22 24/03/2021'
    timed_out: false
    timeout: 15
    username: domain\\john.smith
    uuid: 0b6abd26-6027-49f7-a535-898159728d71
    ```

!!! info "Task Keys"

    __author__: the author of the image the task was started from

    __cancelled_at__: when the task was cancelled (if cancelled)

    __cancelled_by__: the user who cancelled the task (if cancelled)

    __duration__: how long the task ran for (if completed)

    __exit_code__: the task exit code (if completed)

    __exit_message__: the message corresponding to the __exit_code__ (if completed & configured)

    __id__: the task ID

    __image__: the name of the inventory image the task was started from (also known as the task name)

    __image_id__: the ID of the inventory image

    __started_at__: when the task was started

    __state__: the task state

    __stopped_at__: when the task was stopped (if completed)

    __timeout__: the task timeout in minutes

    __username:__ the name of the user who started the task

    __uuid:__ the task UUID
    """

    author: str
    cancelled_at: Optional[DateString]
    cancelled_by: Optional[str]
    duration: Optional[str]
    exit_code: Optional[int]
    exit_message: Optional[str]
    id: int
    image: str
    image_id: int
    started_at: Optional[DateString]
    state: str
    stopped_at: Optional[DateString]
    timed_out: Optional[bool]
    timeout: int
    username: str
    uuid: str


class TaskData(CastDataClass):
    __annotations__ = Task.__annotations__
    __class_config__ = {
        "cast_functions": {"types": {Optional[DateString]: _optional_date_string}, "fields": {"state": lambda x: x.title()}},
        "always_cast": ["cancelled_at", "started_at", "stopped_at", "state"],
    }


class DelayedTask(TypedDict):
    """
- Contains information about a delayed PPA Task

- Delayed Tasks are all based on an Inventory [Image](#image)

!!! example

    ```YAML linenums="1"
    description: Remove John Smith's Temporary Access
    id: 7
    image: Remove User From Groups
    image_id: 11
    is_owner: false
    is_pending: false
    payload:
      groups:
        - Remote Desktop Users
        - Account Operators
    start_time: '2021-03-25T10:33:43.244374'
    task_uuid: 50d10ddc-be19-464d-8d5f-66b17ac38978
    timezone: Etc/UTC
    trigger: api
    username: ad\access.admin
    ```

!!! info "DelayedTask Keys"

    __description__: a description of the delayed task

    __id__: the delayed task ID

    __image__: the name of the inventory image the delayed task will be started from (also known as the task name)

    __image_id__: the ID of the inventory image

    __is_owner__: denotes whether the delayed task was created by the current user

    __is_pending__: denotes whether the delayed task is still pending

    __payload__: the payload supplied to the task (if applicable)

    __start_time__: the time the task will start

    __task_uuid__: the UUID of the task if it (if running or complete)

    __timezone__: the timezone of the start time

    __trigger__: the method the delayed task was created (`api`, `ui`, or `task`)

    __username__: the name of the user who created the delayed task
    """

    description: str
    id: int
    image: str
    image_id: int
    is_owner: bool
    is_pending: bool
    payload: Optional[AnyJson]
    start_time: DateString
    task_uuid: str
    timezone: str
    trigger: str
    username: str


class DelayedTaskData(CastDataClass):
    __annotations__ = DelayedTask.__annotations__
    __class_config__ = {
        "cast_functions": {"types": {DateString: _date_string}},
        "always_cast": ["start_time"],
    }


class TaskResult(TypedDict):
    """
- Contains a condensed view of a [Task's](#task) result once it has completed

- Also includes any **result_json** saved by a task

- The **result_json** key contains information saved by the task when it ran

- This is mainly used to retrieve a task's output when starting tasks via the API

!!! example

    ```YAML linenums="1"
    exit_code: 0
    exit_message:
    result_json:
      audited_servers:
        - server-001
        - server-002
        - server-003
    state: success
    ```

!!! info "TaskResult Keys"

    __exit_code__: the task exit code (if completed)

    __exit_message__: the message corresponding to the __exit_code__ (if completed & configured)

    __result_json__: any JSON data saved by the task

    __state__: the task state
    """

    exit_code: Optional[int]
    exit_message: Optional[str]
    result_json: AnyJson
    state: str


class TaskResultData(CastDataClass):
    __annotations__ = TaskResult.__annotations__


class Image(TypedDict):
    """
- Contains information about a PPA Inventory Image

- These Inventory Images can be started as [Tasks](#task)

!!! example

    ```YAML linenums="1"
    author: John Smith
    description: Check the status of a service on one or more Windows Servers.
    deployed: true
    id: 37
    name: Check Windows Service
    owner: domain\\john.smith
    tags:
    - windows
    - service
    updated_at: 11:05:28 12/11/2020
    uuid: 96979ffb-649d-4f66-a492-8a20a81eb9bc
    ```

!!! info "Image Keys"

    __author__: the author of the image

    __description__: a description of the image

    __deployed__: denotes whether the task is deployed

    __id__: the image ID

    __name__: the name of the image

    __tags__: a list of tags assigned to the image

    __updated_at__: when the image was last updated

    __uuid__: the image UUID
    """

    author: str
    description: Optional[str]
    deployed: bool
    id: int
    name: str
    owner: str
    tags: List[str]
    updated_at: DateString
    uuid: str


class ImageData(CastDataClass):
    __annotations__ = Image.__annotations__
    __class_config__ = {
        "cast_functions": {"types": {DateString: _date_string}, "fields": {"deployed": lambda x: bool(x)}},
        "always_cast": ["updated_at", "deployed"],
    }


class TimeStamp(TypedDict):
    """
- Contains submitted date picker information

!!! info "All TimeStamp Keys"

    __day__: The selected day.

    __month__: The selected month.

    __year__: The selected year.

    __hour__: The selected hour.

    __minute__: The selected minute.

    __second__: The submitted second.

    __timezone__: The chosen timezone.

    __timestamp__: The supplied date, time, & UTC offset in ISO-8601 format (`yyyy-mm-ddThh:mm:ss+hh:mm`).
    """

    hour: int
    minute: int
    second: int
    day: int
    month: int
    year: int
    timezone: str
    timestamp: str
