"""
Tasks

This module contains actions for starting tasks in PPA & viewing task history.

!!! danger "PPA Version"

    The actions in this module require PPA v2.7.x or newer.
"""
import math
import uuid
import logging
import datetime

from typing import Dict, Any, Optional, List, Union

from dateutil import parser

from ppa import ui

from ._api import get_client, create_records, get_cell_value, min_ppa_version
from .records import Task, DelayedTask, TaskResult, TimeStamp


logger = logging.getLogger(__name__)

TABLE_HEADER = ["Name", "Started By", "Started At", "Stopped At", "State"]
TABLE_FIELDS = ["image", "username", "started_at", "stopped_at", "state"]


@create_records(Task)
def get_started_by_me() -> List[Task]:
    """Get all tasks that were started by the current user.

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    This action takes no inputs.

??? success "Output"

    A list of [Task](../inputs_outputs/#task) dictionaries.

??? example
    ```YAML linenums="1"
    - ppa_api.tasks.get_started_by_me:
      save: my_tasks
    ```
"""
    with get_client() as ppa:
        return ppa.tasks_started_by_me()


@create_records(Task)
def get_all() -> List[Task]:
    """Get all tasks visible to the current user.

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    This action takes no inputs.

??? success "Output"

    A list of [Task](../inputs_outputs/#task) dictionaries.

??? example
    ```YAML linenums="1"
    - ppa_api.tasks.get_all:
      save: all_tasks
    ```
"""
    with get_client() as ppa:
        return ppa.tasks()


@create_records(TaskResult)
def get_result(uuid: str) -> TaskResult:
    """Get the saved result from the supplied task.

This is the only action that outputs the `result_json` saved by a task.

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    - __uuid:__ the UUID of the task

??? success "Output"

    A single [TaskResult](../inputs_outputs/#taskresult) dictionary.

??? example
    ```YAML linenums="1"
    - ppa_api.tasks.get_result:
      load:
        uuid: task.uuid
      save: task_result
    ```
"""
    with get_client() as ppa:
        return ppa.get_task_result(uuid)


@min_ppa_version("2.7.1")
@create_records(Task)
def start(name: str, payload: Optional[Dict[str, Any]] = None, timeout: Optional[int] = 600) -> Task:
    """Start a task & wait for it to complete.

!!! question "Minimum Plugin Version: 2.0.0"

!!! warning "Task Interaction"

    This plugin cannot interact with a task once it has started.

    If the task requires user input once started, the task may never complete.

    You can use payloads to create tasks that require no interaction.

??? info "Input"

    - __name:__ the name of the inventory image

    - __payload:__ an optional payload to supply to the task (defaults to empty)

    - __timeout:__ optional number of seconds to wait (defaults to `600`)

??? success "Output"

    A single [Task](../inputs_outputs/#task) dictionary.

??? example
    ```YAML linenums="1"
    - ppa_api.tasks.start:
        name: Local Group Audit
        payload:
          group_name: Remote Desktop Users
      save: task
    ```
"""
    with get_client() as ppa:
        return ppa.start_task(name, payload=payload, timeout=timeout)


@min_ppa_version("2.8.0")
@create_records(DelayedTask)
def start_after(name: str, description: str, minutes: int, payload: Optional[Dict[str, Any]] = None) -> DelayedTask:
    """Create a delayed task that will start after a number of minutes has elapsed.

!!! question "Minimum Plugin Version: 2.0.0"

!!! question "Minimum PPA Version: 2.8.0"

??? info "Input"

    - __name:__ the name of the inventory image

    - __description:__ a description explaining the purpose of the delayed task

    - __minutes:__ the number of minutes to wait until starting the task

    - __payload:__ an optional payload to supply to the task (defaults to empty)

??? success "Output"

    A single [DelayedTask](../inputs_outputs/#delayedtask) dictionary.

??? example
    ```YAML linenums="1"
    - ppa_api.delayed_tasks.create:
        name: Local Group Audit
        delay: 60
        payload:
          group_name: Remote Desktop Users
      save: delayed_task
    ```
"""
    with get_client() as ppa:
        return ppa.delay_task(name, delay=minutes, description=description, payload=payload)


@min_ppa_version("2.8.0")
def start_at(name: str, description: str, timestamp: TimeStamp, payload: Optional[Dict[str, Any]] = None) -> DelayedTask:
    """Create a delayed task that will start at a certain time & date.

!!! question "Minimum Plugin Version: 2.0.0"

!!! question "Minimum PPA Version: 2.8.0"

??? info "Input"

    - __name:__ the name of the inventory image

    - __description:__ a description explaining the purpose of the delayed task

    - __timestamp:__ a [TimeStamp](../inputs_outputs/#timestamp) outputted by the [input_timestamp](/plugins/ppa/ppa.ui/#input_timestamp) action

    - __payload:__ an optional payload to supply to the task (defaults to empty)

??? success "Output"

    A single [DelayedTask](../inputs_outputs/#delayedtask) dictionary.

??? example
    ```YAML linenums="1"
    - ppa.ui.input_timestamp:
        text: Supply a Start Time
      save: start_time

    - ppa_api.delayed_tasks.create:
        name: Local Group Audit
        payload:
          group_name: Remote Desktop Users
      load:
        timestamp: start_time
      save: delayed_task
    ```
"""
    # We don't need the timestamp string, only the timezone string.
    logger.info(f"timestamp: {timestamp}")
    supplied_timestamp, timezone = list(map(lambda key: timestamp.pop(key), ("timestamp", "timezone")))

    start_local = parser.isoparse(supplied_timestamp)
    logger.info(f"supplied start time: {start_local}")
    start_utc = start_local.astimezone(datetime.timezone.utc)
    logger.info(f"start time in UTC: {start_utc}")

    now_utc = datetime.datetime.now().astimezone(datetime.timezone.utc)
    logger.info(f"now in UTC: {now_utc}")

    # Now work out the difference between then & now in minutes, but round it **up** to the nearest minute.
    delay = math.ceil((start_utc - now_utc).total_seconds() / 60)

    logger.info(f"there are {delay} minutes between now UTC ({now_utc}) & the start time in UTC ({start_utc})")
    with get_client() as ppa:
        return ppa.delay_task(name, delay=delay, description=description, payload=payload)


@min_ppa_version("2.7.1")
@create_records(Task)
def start_async(name: str, payload: Optional[Dict[str, Any]] = None) -> Task:
    """Start a task & continue without waiting.

Use this action if:

- You don't need to wait for the task to finish

- You want to loop & poll the task state yourself

!!! question "Minimum Plugin Version: 2.0.0"

!!! warning "Task Interaction"

    This plugin cannot interact with a task once it has started.

    If the task requires user input once started, the task may never complete.

    You can use payloads to create tasks that require no interaction.

??? info "Input"

    - __name:__ the name of the inventory image

    - __payload:__ an optional payload to supply to the task (defaults to empty)

??? success "Output"

    A single [Task](../inputs_outputs/#task) dictionary.

??? example
    ```YAML linenums="1"
    - ppa_api.tasks.start_async:
        name: Local Group Audit
        payload:
          group_name: Remote Desktop Users
      save: task
    ```
"""
    with get_client() as ppa:
        return ppa.start_task_async(name, payload=payload)


def cancel(uuid: str) -> None:
    """Cancel a running task.

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    - __uuid:__ the task UUID

??? success "Output"

    Nothing is outputted by this action.

??? example
    ```YAML linenums="1"
    - ppa_api.tasks.start_async:
        name: Local Group Audit
        payload:
          group_name: Remote Desktop Users
      save: task

    - ppa_api.tasks.cancel:
      load:
        uuid: task.uuid
    ```
"""
    with get_client() as ppa:
        return ppa.cancel_task(uuid)


def output_table(text: str, tasks: Union[List[Task], Task]) -> None:
    """Display tasks in a table.

The table will have the following columns:

- Name
- Started By
- Started At
- Stopped At
- State

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __tasks:__ a single or list of [Task](../inputs_outputs/#task) dictionaries to display in the table

??? success "Output"

    Nothing is outputted by this action.

??? example
    ```YAML linenums="1"
    - ppa_api.tasks.started_by_me:
      save: my_tasks

    - ppa_api.tasks.output_table:
        text: My Tasks
      load:
        tasks: my_tasks
    ```
    """
    tasks = [tasks] if not isinstance(tasks, list) else tasks
    ui.output_table(text, header=TABLE_HEADER, data=[[get_cell_value(task, field) for field in TABLE_FIELDS] for task in tasks])


def input_table(
    text: str, tasks: Union[List[Task], Task], minimum: Optional[int] = None, maximum: Optional[int] = None
) -> List[Task]:
    """Display tasks in a table, & allow the task operator to make a selection.

The table will have the following columns:

- Name
- Started By
- Started At
- Stopped At
- State

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __tasks:__ a single or list of [Task](../inputs_outputs/#task) dictionaries to display in the table

    - __minimum:__ the minimum number of acceptable selections

    - __maximum:__ the maximum number of acceptable selections

??? success "Output"

    A single or list of [Task](../inputs_outputs/#task) dictionaries.

??? tip

    - If neither a _minimum_ or _maximum_ is provided, the task operator will be able to submit 0 selections.
    - If _minimum_ or _maximum_ are provided, the operation will repeat until the task operator makes a valid number
        of selections.

??? example

    ```YAML linenums="1"
    - ppa_api.tasks.started_by_me:
      save: my_tasks

    - ppa_api.tasks.input_table:
        text: My Tasks
        minimum: 1
        maximum: 1
      load:
        tasks: my_tasks
      save: selected_task

    - ppa_api.tasks.get_result:
      load:
        uuid: selected_task.uuid
    ```
    """
    tasks = [tasks] if not isinstance(tasks, list) else tasks
    table_element_id = str(uuid.uuid4())
    error_element_id = str(uuid.uuid4())
    while True:
        indices = ui.input_table(
            text,
            header=TABLE_HEADER,
            data=[[get_cell_value(task, field) for field in TABLE_FIELDS] for task in tasks],
            element_id=table_element_id,
        )
        count = len(indices)
        if minimum and count < minimum:
            ui.output_error(f"Please choose at least {minimum} record(s)", element_id=error_element_id)
            continue
        if maximum and count > maximum:
            ui.output_error(f"Please choose no more than {maximum} record(s)", element_id=error_element_id)
            continue
        break

    return [tasks[i] for i in indices]
